// Copyright 2023 Azul Systems, Inc.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

package jdk.internal.crac;

import jdk.internal.crac.mirror.Context;
import jdk.internal.crac.mirror.Resource;
import sun.security.action.GetBooleanAction;

import java.util.function.Supplier;

public abstract class JDKFdResource implements JDKResource {
    public static final String COLLECT_FD_STACKTRACES_PROPERTY = "jdk.crac.collect-fd-stacktraces";
    private static final String COLLECT_FD_STACKTRACES_HINT =
        "Use -D" + COLLECT_FD_STACKTRACES_PROPERTY + "=true to find the source.";

    private static final boolean COLLECT_FD_STACKTRACES =
        GetBooleanAction.privilegedGetProperty(COLLECT_FD_STACKTRACES_PROPERTY);

    // No lambdas during clinit...
    protected static Supplier<Exception> NO_EXCEPTION = new Supplier<Exception>() {
        @Override
        public Exception get() {
            return null;
        }
    };

    final Exception stackTraceHolder;

    static volatile boolean stacktraceHintPrinted = false;
    static volatile boolean warningSuppressionHintPrinted = false;

    public JDKFdResource() {
        stackTraceHolder = COLLECT_FD_STACKTRACES ?
            // About the timestamp: we cannot format it nicely since this
            // exception is sometimes created too early in the VM lifecycle
            // (but it's hard to detect when it would be safe to do).
            new Exception("This file descriptor was created by " + Thread.currentThread().getName()
                + " at epoch:" + System.currentTimeMillis() + " here") :
            null;

        Core.Priority.FILE_DESCRIPTORS.getContext().register(this);
        OpenResourcePolicies.ensureRegistered();
    }

    protected Exception getStackTraceHolder() {
        if (!stacktraceHintPrinted && stackTraceHolder == null) {
            stacktraceHintPrinted = true;
            LoggerContainer.info(COLLECT_FD_STACKTRACES_HINT);
        }
        return stackTraceHolder;
    }

    @Override
    public void afterRestore(Context<? extends Resource> context) throws Exception {

    }

    protected void warnOpenResource(OpenResourcePolicies.Policy policy, String self) {
        // The warning is not printed for implicitly closed resource (without policy)
        // e.g. standard input/output streams
        String warn = "false";
        if (policy != null) {
            warn = policy.params.getOrDefault("warn", "true");
        }
        if (Boolean.parseBoolean(warn)) {
            LoggerContainer.warn("{0} was not closed by the application.", self);
            if (!warningSuppressionHintPrinted) {
                LoggerContainer.info("To suppress the warning above use 'warn: false' in the resource policy.");
                warningSuppressionHintPrinted = true;
            }
        }
    }
}
